import functools
import locale
import os
import sys
import unittest

from io import StringIO
from unittest import skipUnless, skipIf, expectedFailure

import blessed
import pyte
from pyte import control as ctrl, Stream, Screen

from curtsies.window import BaseWindow, FullscreenWindow, CursorAwareWindow


class FakeStdin(StringIO):
    encoding = "ascii"


# thanks superbobry for this code: https://github.com/selectel/pyte/issues/13
class ReportingStream(Stream):
    report_escape = {"6": "report_cursor_position"}

    def _arguments(self, char):
        if char == "n":
            # DSR command looks like 'CSI<N>n'. So all we need to do
            # is wait for the 'n' argument.
            return self.dispatch(self.report_escape[self.current])
        else:
            return super()._arguments(char)


class ReportingScreen(Screen):
    def __init__(self, *args, **kwargs):
        self._report_file = FakeStdin()
        super().__init__(*args, **kwargs)

    def report_cursor_position(self):
        # cursor position is 1-indexed in the ANSI escape sequence API
        s = ctrl.CSI + "%d;%sR" % (self.cursor.y + 1, self.cursor.x + 1)
        self._report_file.seek(0)
        self._report_file.write(s)
        self._report_file.seek(0)


class ReportingScreenWithExtra(ReportingScreen):
    def report_cursor_position(self):
        # cursor position is 1-indexed in the ANSI escape sequence API
        extra = "qwerty\nasdf\nzxcv"
        s = ctrl.CSI + "%d;%sR" % (self.cursor.y + 1, self.cursor.x + 1)
        self._report_file.seek(0)
        self._report_file.write(extra + s)
        self._report_file.seek(0)


class Bugger:
    __before__ = __after__ = lambda *args: None

    def __getattr__(self, event):
        to = sys.stdout

        def inner(*args, **flags):
            to.write(event.upper() + " ")
            to.write("; ".join(map(repr, args)))
            to.write(" ")
            to.write(
                ", ".join(
                    "{}: {}".format(name, repr(arg)) for name, arg in flags.items()
                )
            )
            to.write(os.linesep)

        return inner


class ScreenStdout:
    def __init__(self, stream):
        self.stream = stream

    def write(self, s):
        if sys.version_info[0] == 3:
            self.stream.feed(s)
        else:
            self.stream.feed(s.decode(locale.getpreferredencoding()))

    def flush(self):
        pass


@skipUnless(sys.stdin.isatty(), "blessed Terminal needs streams open")
class TestFullscreenWindow(unittest.TestCase):
    def setUp(self):
        self.screen = pyte.Screen(10, 3)
        self.stream = pyte.Stream()
        self.stream.attach(self.screen)
        stdout = ScreenStdout(self.stream)
        self.window = FullscreenWindow(stdout)

    def test_render(self):
        with self.window:
            self.window.render_to_terminal(["hi", "there"])
        self.assertEqual(
            self.screen.display, ["hi        ", "there     ", "          "]
        )

    def test_scroll(self):
        with self.window:
            self.window.render_to_terminal(["hi", "there"])
            self.window.scroll_down()
        self.assertEqual(
            self.screen.display, ["there     ", "          ", "          "]
        )


class NopContext:
    def __enter__(*args):
        pass

    def __exit__(*args):
        pass


@skipUnless(sys.stdin.isatty(), "blessed Terminal needs streams open")
class TestCursorAwareWindow(unittest.TestCase):
    def setUp(self):
        self.screen = ReportingScreen(6, 3)
        self.stream = ReportingStream()
        self.stream.attach(self.screen)
        self.stream.attach(Bugger())
        stdout = ScreenStdout(self.stream)
        self.window = CursorAwareWindow(
            out_stream=stdout, in_stream=self.screen._report_file
        )
        self.window.cbreak = NopContext()
        blessed.Terminal.height = 3
        blessed.Terminal.width = 6

    # This isn't passing locally for me anymore :/
    @expectedFailure
    def test_render(self):
        with self.window:
            self.assertEqual(self.window.top_usable_row, 0)
            self.window.render_to_terminal(["hi", "there"])
            self.assertEqual(self.screen.display, ["hi    ", "there ", "      "])

    # This isn't passing locally for me anymore :/
    @expectedFailure
    def test_cursor_position(self):
        with self.window:
            self.window.render_to_terminal(["hi", "there"], cursor_pos=(2, 4))
            self.assertEqual(self.window.get_cursor_position(), (2, 4))

    # This isn't passing locally for me anymore :/
    @expectedFailure
    def test_inital_cursor_position(self):

        self.screen.cursor.y += 1
        with self.window:
            self.assertEqual(self.window.top_usable_row, 1)
            self.window.render_to_terminal(["hi", "there"])
            self.assertEqual(self.screen.display, ["      ", "hi    ", "there "])


@skipUnless(sys.stdin.isatty(), "blessed Terminal needs streams open")
class TestCursorAwareWindowWithExtraInput(unittest.TestCase):
    def setUp(self):
        self.screen = ReportingScreenWithExtra(6, 3)
        self.stream = ReportingStream()
        self.stream.attach(self.screen)
        self.stream.attach(Bugger())
        stdout = ScreenStdout(self.stream)
        self.extra_bytes = []
        self.window = CursorAwareWindow(
            out_stream=stdout,
            in_stream=self.screen._report_file,
            extra_bytes_callback=self.extra_bytes_callback,
        )
        self.window.cbreak = NopContext()
        blessed.Terminal.height = 3
        blessed.Terminal.width = 6

    def extra_bytes_callback(self, bytes):
        self.extra_bytes.append(bytes)

    # This isn't passing locally for me anymore :/
    @expectedFailure
    def test_report_extra_bytes(self):
        with self.window:
            pass  # should have triggered getting initial cursor position
        self.assertEqual(b"".join(self.extra_bytes), b"qwerty\nasdf\nzxcv")



import unittest
import sys

from curtsies.window import BaseWindow, FullscreenWindow, CursorAwareWindow
from io import StringIO
from unittest import skipIf


fds_closed = not sys.stdin.isatty() or not sys.stdout.isatty()


class FakeFullscreenWindow(FullscreenWindow):
    width = property(lambda self: 10)
    height = property(lambda self: 4)


@skipIf(fds_closed, "blessed Terminal needs streams open")
class TestBaseWindow(unittest.TestCase):
    """Pretty pathetic tests for window"""

    def test_window(self):
        fakestdout = StringIO()
        window = BaseWindow(fakestdout)
        window.write("hi")
        fakestdout.seek(0)
        self.assertEqual(fakestdout.read(), "hi")

    def test_array_from_text(self):
        window = BaseWindow()
        a = window.array_from_text(".\n.\n.")
        self.assertEqual(a.height, 3)
        self.assertEqual(a[0], ".")
        self.assertEqual(a[1], ".")

    def test_array_from_text_rc(self):
        a = BaseWindow.array_from_text_rc("asdfe\nzx\n\n123", 3, 4)
        self.assertEqual(a.height, 3)
        self.assertEqual(a.width, 4)
        self.assertEqual(a[0], "asdf")
        self.assertEqual(a[1], "e")
        self.assertEqual(a[2], "zx")

    def test_fullscreen_window(self):
        fakestdout = StringIO()
        window = FullscreenWindow(fakestdout)
        window.write("hi")
        fakestdout.seek(0)
        self.assertEqual(fakestdout.read(), "hi")

    def test_fullscreen_render_to_terminal(self):
        fakestdout = StringIO()
        window = FakeFullscreenWindow(fakestdout)
        window.render_to_terminal(["hello", "hello", "hello"])
        fakestdout.seek(0)
        output = fakestdout.read()
        self.assertEqual(output.count("hello"), 3)



import time

from curtsies import FullscreenWindow, Input, FSArray
from curtsies.fmtfuncs import red, bold, green, on_blue, yellow, on_red
import curtsies.events

class Frame(curtsies.events.ScheduledEvent):
    pass

class World:
    def __init__(self):
        self.s = 'Hello'
    def tick(self):
        self.s += '|'
        self.s = self.s[max(1, len(self.s)-80):]
    def process_event(self, e):
        self.s += str(e)

def realtime(fps=15):
    world = World()
    dt = 1/fps

    reactor = Input()
    schedule_next_frame = reactor.scheduled_event_trigger(Frame)
    schedule_next_frame(when=time.time())

    with reactor:
        for e in reactor:
            if isinstance(e, Frame):
                world.tick()
                print(world.s)
                when = e.when + dt
                while when < time.time():
                    when += dt
                schedule_next_frame(when)
            elif e == '<ESC>':
                break
            else:
                world.process_event(e)

if __name__ == "__main__":
    realtime()



from setuptools import setup
import ast
import os


def version():
    """Return version string."""
    with open(os.path.join("curtsies", "__init__.py")) as input_file:
        for line in input_file:
            if line.startswith("__version__"):
                return ast.parse(line).body[0].value.s


setup(
    version=version(),
)



import time
import sys

from curtsies import FullscreenWindow, Input, FSArray, fmtstr, fsarray
from curtsies.fmtfuncs import (
    bold,
    yellow,
    on_blue,
    cyan,
    on_yellow,
    on_red,
    black,
)
import curtsies.events


class Cycle:
    def __init__(self, attr_dict):
        self.appearance = attr_dict["appearance"]
        self.x, self.y = attr_dict["x"], attr_dict["y"]
        self.keylist = attr_dict["keys"]
        self.dir = 0

    def move(self, grid):
        d = self.dir
        if d == 0:
            self.x += 1
        elif d == 90:
            self.y -= 1
        elif d == 180:
            self.x -= 1
        elif d == 270:
            self.y += 1

    def face(self, newdir):
        """ turn to face the given direction """
        if not (newdir % 180 == self.dir % 180):
            self.dir = newdir

    def paint(self, grid):
        """ given a grid object, adds self to the grid and returns new grid
        """
        grid[self.y, self.x] = self.appearance
        return grid


class Bot(Cycle):
    """ Dumb bot that only turns right when it's about to crash
    """

    def move(self, grid):

        a = self.getNextSquareCoords()
        if grid[a[0], a[1]] != [" "]:
            self.turnLeft()
        super().move(grid)

    def turnRight(self):
        self.dir = (self.dir - 90) % 360

    def turnLeft(self):
        self.dir = (self.dir + 90) % 360

    def getNextSquareCoords(self, n=2):
        """ Returns the x, y coords of the box n spots in front of the bot.
        """
        if self.dir == 0:
            return (self.y, self.x + n)
        elif self.dir == 90:
            return (self.y - n, self.x)
        elif self.dir == 180:
            return (self.y, self.x - n)
        else:
            return (self.y + n, self.x)


class gameboard:
    def __init__(self, width, height, players):
        self.width = width
        self.height = height
        self.grid = FSArray(height, width)

        self.players = players
        self.numplayers = len(self.players)

    def draw_border(self):
        """ draw outer border """
        box = on_yellow(black(bold("!")))
        for x in range(1, self.width):
            self.grid[1, x] = box
            self.grid[self.height - 1, x] = box
        for y in range(1, self.height):
            self.grid[y, 1] = box
            self.grid[y, self.width - 1] = box

    def process_event(self, key):
        if key == " ":
            sys.exit()

        for player in self.players:
            if key in player.keylist:
                player.face(player.keylist[key])
                return False

    def tick(self):
        """ Do one frame of work. Returns the winner if there 
        is a crash
        """
        # list of players alive this frame
        temp_players = self.players[:]

        for bike in self.players:
            bike.move(self.grid)

            if self.grid[bike.y, bike.x] == [" "]:
                self.grid[bike.y, bike.x] = bike.appearance
            else:  # crashed
                temp_players.remove(bike)
                self.grid[bike.y, bike.x] = fmtstr("X", "red", "bold", "on_black")

        if len(temp_players) == 0:
            return "tie"
        elif len(temp_players) < len(self.players):
            return temp_players

    def winner_msg(self, tick):
        if len(tick) != 1:
            msg = fmtstr("it's a tie!", "red")
        else:
            winner = tick[0]
            winner_name = winner.appearance
            msg = fmtstr("winner:", "yellow")
            msg += " " + (winner_name * 5)
        return msg


class Frame(curtsies.events.ScheduledEvent):
    pass


def do_introduction(window):
    h, w = window.height, window.width

    messages = [
            "two player tron",
            fmtstr("player 1:", "on_blue", "cyan") + " wasd",
            fmtstr("player 2:", "on_red", "yellow") + " arrow keys",
        ]

    billboard = FSArray(h, w)
    msg_row = h // 2 - 2
    for msg in messages:
        billboard[
            msg_row, w // 2 - len(msg) // 2 : w // 2 + len(msg) // 2 + 1
        ] = fsarray([msg])
        msg_row += 1
    window.render_to_terminal(billboard)

    # countdown msg
    for i in range(3, 0, -1):
        billboard[msg_row, w // 2] = fmtstr(str(i), "red")
        window.render_to_terminal(billboard)
        time.sleep(1)


def mainloop(window, p2_bot=False):
    p1_attrs = {
        "appearance": on_blue(cyan("1")),
        "x": window.width // 4,
        "y": window.height // 2,
        "keys": {"w": 90, "a": 180, "s": 270, "d": 0},
    }

    p2_attrs = {
        "appearance": on_red(yellow("2")),
        "x": 3 * window.width // 4,
        "y": window.height // 2,
        "keys": {"<UP>": 90, "<LEFT>": 180, "<DOWN>": 270, "<RIGHT>": 0},
    }

    FPS = 15

    players = [Cycle(p1_attrs), Cycle(p2_attrs)]
    if p2_bot: # make p2 a bot
        players[1] = Bot(p2_attrs)

    world = gameboard(window.width, window.height, players)
    dt = 1 / FPS
    world.draw_border()
    window.render_to_terminal(world.grid)

    reactor = Input()
    schedule_next_frame = reactor.scheduled_event_trigger(Frame)
    schedule_next_frame(when=time.time())
    with reactor:
        for c in reactor:
            if isinstance(c, Frame):
                tick = world.tick()
                window.render_to_terminal(world.grid)
                if not tick:  # if no crashes
                    when = c.when + dt
                    while when < time.time():
                        when += dt
                    schedule_next_frame(when)
                else:  # if crashed
                    world.grid[0:4, 0:25] = fsarray(
                        [
                            world.winner_msg(tick),
                            "r to restart",
                            "q to quit",
                            "b to make player 2 a bot",
                        ]
                    )
                    window.render_to_terminal(world.grid)
            elif c.lower() in ["r", "q", "b"]:
                break
            else:  # common case
                world.process_event(c)
    if c.lower() == "r":
        mainloop(window, p2_bot)
    elif c.lower() == "b":
        mainloop(window, True)


def main():
    with FullscreenWindow(sys.stdout) as window:
        do_introduction(window)

        mainloop(window)


if __name__ == "__main__":
    main()